@@ -13,6 +13,7 @@ import codecs
 import mimetypes
 from copy import deepcopy
 from itertools import repeat
+from collections import Container, Iterable, MutableSet
 
 from werkzeug._internal import _missing, _empty_stream
 from werkzeug._compat import iterkeys, itervalues, iteritems, iterlists, \
@@ -51,7 +52,14 @@ def native_itermethods(names):
     if not PY2:
         return lambda x: x
 
-    def setmethod(cls, name):
+    def setviewmethod(cls, name):
+        viewmethod_name = 'view%s' % name
+        viewmethod = lambda self, *a, **kw: ViewItems(self, name, 'view_%s' % name, *a, **kw)
+        viewmethod.__doc__ = \
+            '"""`%s()` object providing a view on %s"""' % (viewmethod_name, name)
+        setattr(cls, viewmethod_name, viewmethod)
+
+    def setitermethod(cls, name):
         itermethod = getattr(cls, name)
         setattr(cls, 'iter%s' % name, itermethod)
         listmethod = lambda self, *a, **kw: list(itermethod(self, *a, **kw))
@@ -61,7 +69,8 @@ def native_itermethods(names):
 
     def wrap(cls):
         for name in names:
-            setmethod(cls, name)
+            setitermethod(cls, name)
+            setviewmethod(cls, name)
         return cls
     return wrap
 
@@ -89,9 +98,6 @@ class ImmutableListMixin(object):
     def __delitem__(self, key):
         is_immutable(self)
 
-    def __delslice__(self, i, j):
-        is_immutable(self)
-
     def __iadd__(self, other):
         is_immutable(self)
     __imul__ = __iadd__
@@ -99,9 +105,6 @@ class ImmutableListMixin(object):
     def __setitem__(self, key, value):
         is_immutable(self)
 
-    def __setslice__(self, i, j, value):
-        is_immutable(self)
-
     def append(self, item):
         is_immutable(self)
     remove = append
@@ -322,6 +325,25 @@ class ImmutableTypeConversionDict(ImmutableDictMixin, TypeConversionDict):
         return self
 
 
+class ViewItems(object):
+
+    def __init__(self, multi_dict, method, repr_name, *a, **kw):
+        self.__multi_dict = multi_dict
+        self.__method = method
+        self.__repr_name = repr_name
+        self.__a = a
+        self.__kw = kw
+
+    def __get_items(self):
+        return getattr(self.__multi_dict, self.__method)(*self.__a, **self.__kw)
+
+    def __repr__(self):
+        return '%s(%r)' % (self.__repr_name, list(self.__get_items()))
+
+    def __iter__(self):
+        return iter(self.__get_items())
+
+
 @native_itermethods(['keys', 'values', 'items', 'lists', 'listvalues'])
 class MultiDict(TypeConversionDict):
 
@@ -372,6 +394,8 @@ class MultiDict(TypeConversionDict):
             tmp = {}
             for key, value in iteritems(mapping):
                 if isinstance(value, (tuple, list)):
+                    if len(value) == 0:
+                        continue
                     value = list(value)
                 else:
                     value = [value]
@@ -398,7 +422,9 @@ class MultiDict(TypeConversionDict):
         :raise KeyError: if the key does not exist.
         """
         if key in self:
-            return dict.__getitem__(self, key)[0]
+            lst = dict.__getitem__(self, key)
+            if len(lst) > 0:
+                return lst[0]
         raise exceptions.BadRequestKeyError(key)
 
     def __setitem__(self, key, value):
@@ -602,7 +628,12 @@ class MultiDict(TypeConversionDict):
                         not in the dictionary.
         """
         try:
-            return dict.pop(self, key)[0]
+            lst = dict.pop(self, key)
+
+            if len(lst) == 0:
+                raise exceptions.BadRequestKeyError()
+
+            return lst[0]
         except KeyError as e:
             if default is not _missing:
                 return default
@@ -612,6 +643,10 @@ class MultiDict(TypeConversionDict):
         """Pop an item from the dict."""
         try:
             item = dict.popitem(self)
+
+            if len(item[1]) == 0:
+                raise exceptions.BadRequestKeyError()
+
             return (item[0], item[1][0])
         except KeyError as e:
             raise exceptions.BadRequestKeyError(str(e))
@@ -724,6 +759,8 @@ class OrderedMultiDict(MultiDict):
                 return False
         return True
 
+    __hash__ = None
+
     def __ne__(self, other):
         return not self.__eq__(other)
 
@@ -931,6 +968,8 @@ class Headers(object):
         return other.__class__ is self.__class__ and \
             set(other._list) == set(self._list)
 
+    __hash__ = None
+
     def __ne__(self, other):
         return not self.__eq__(other)
 
@@ -1299,6 +1338,8 @@ class EnvironHeaders(ImmutableHeadersMixin, Headers):
     def __eq__(self, other):
         return self.environ is other.environ
 
+    __hash__ = None
+
     def __getitem__(self, key, _get_mode=False):
         # _get_mode is a no-op for this class as there is no index but
         # used because get() calls it.
@@ -1596,10 +1637,8 @@ class Accept(ImmutableList):
             list.__init__(self, values)
         else:
             self.provided = True
-            values = [(a, b) for b, a in values]
-            values.sort()
-            values.reverse()
-            list.__init__(self, [(a, b) for b, a in values])
+            values = sorted(values, key=lambda x: (x[1], x[0]), reverse=True)
+            list.__init__(self, values)
 
     def _value_matches(self, value, item):
         """Check if a value matches a given accept item."""
@@ -1955,7 +1994,7 @@ class CallbackDict(UpdateDictMixin, dict):
         )
 
 
-class HeaderSet(object):
+class HeaderSet(MutableSet):
 
     """Similar to the :class:`ETags` class this implements a set-like structure.
     Unlike :class:`ETags` this is case insensitive and used for vary, allow, and
@@ -2109,7 +2148,7 @@ class HeaderSet(object):
         )
 
 
-class ETags(object):
+class ETags(Container, Iterable):
 
     """A set that can be used to check if one etag is present in a collection
     of etags.
@@ -2271,6 +2310,17 @@ class Range(object):
                 ranges.append('%s-%s' % (begin, end - 1))
         return '%s=%s' % (self.units, ','.join(ranges))
 
+    def to_content_range_header(self, length):
+        """Converts the object into `Content-Range` HTTP header,
+        based on given length
+        """
+        range_for_length = self.range_for_length(length)
+        if range_for_length is not None:
+            return '%s %d-%d/%d' % (self.units,
+                                    range_for_length[0],
+                                    range_for_length[1] - 1, length)
+        return None
+
     def __str__(self):
         return self.to_header()
 
@@ -2673,7 +2723,7 @@ class FileStorage(object):
         return getattr(self.stream, name)
 
     def __iter__(self):
-        return iter(self.readline, '')
+        return iter(self.stream)
 
     def __repr__(self):
         return '<%s: %r (%r)>' % (
