@@ -23,9 +23,11 @@ try:
 except ImportError:  # pragma: no cover
     from email.Utils import parsedate_tz
 try:
-    from urllib2 import parse_http_list as _parse_list_header
-except ImportError:  # pragma: no cover
     from urllib.request import parse_http_list as _parse_list_header
+    from urllib.parse import unquote_to_bytes as _unquote
+except ImportError:  # pragma: no cover
+    from urllib2 import parse_http_list as _parse_list_header, \
+        unquote as _unquote
 from datetime import datetime, timedelta
 from hashlib import md5
 import base64
@@ -61,7 +63,8 @@ _etag_re = re.compile(r'([Ww]/)?(?:"(.*?)"|(.*?))(?:\s*,\s*|$)')
 _unsafe_header_chars = set('()<>@,;:\"/[]?={} \t')
 _quoted_string_re = r'"[^"\\]*(?:\\.[^"\\]*)*"'
 _option_header_piece_re = re.compile(
-    r';\s*(%s|[^\s;,=]+)\s*(?:=\s*(%s|[^;,]+)?)?\s*' %
+    r';\s*(%s|[^\s;,=\*]+)\s*'
+    r'(?:\*?=\s*(?:([^\s]+?)\'([^\s]*?)\')?(%s|[^;,]+)?)?\s*' %
     (_quoted_string_re, _quoted_string_re)
 )
 _option_header_start_mime_type = re.compile(r',\s*([^;,\s]+)([;,]\s*.+)?')
@@ -125,6 +128,7 @@ HTTP_STATUS_CODES = {
     429:    'Too Many Requests',
     431:    'Request Header Fields Too Large',
     449:    'Retry With',  # proprietary MS extension
+    451:    'Unavailable For Legal Reasons',
     500:    'Internal Server Error',
     501:    'Not Implemented',
     502:    'Bad Gateway',
@@ -354,12 +358,14 @@ def parse_options_header(value, multiple=False):
             optmatch = _option_header_piece_re.match(rest)
             if not optmatch:
                 break
-            option, option_value = optmatch.groups()
+            option, encoding, _, option_value = optmatch.groups()
             option = unquote_header_value(option)
             if option_value is not None:
                 option_value = unquote_header_value(
                     option_value,
                     option == 'filename')
+                if encoding is not None:
+                    option_value = _unquote(option_value).decode(encoding)
             options[option] = option_value
             rest = rest[optmatch.end():]
         result.append(options)
@@ -551,15 +557,24 @@ def parse_range_header(value, make_inclusive=True):
         if item.startswith('-'):
             if last_end < 0:
                 return None
-            begin = int(item)
+            try:
+                begin = int(item)
+            except ValueError:
+                return None
             end = None
             last_end = -1
         elif '-' in item:
             begin, end = item.split('-', 1)
+            begin = begin.strip()
+            end = end.strip()
+            if not begin.isdigit():
+                return None
             begin = int(begin)
             if begin < last_end or last_end < 0:
                 return None
             if end:
+                if not end.isdigit():
+                    return None
                 end = int(end) + 1
                 if begin >= end:
                     return None
@@ -766,7 +781,8 @@ def http_date(timestamp=None):
     return _dump_date(timestamp, ' ')
 
 
-def is_resource_modified(environ, etag=None, data=None, last_modified=None):
+def is_resource_modified(environ, etag=None, data=None, last_modified=None,
+                         ignore_if_range=True):
     """Convenience method for conditional requests.
 
     :param environ: the WSGI environment of the request to be checked.
@@ -774,6 +790,8 @@ def is_resource_modified(environ, etag=None, data=None, last_modified=None):
     :param data: or alternatively the data of the response to automatically
                  generate an etag using :func:`generate_etag`.
     :param last_modified: an optional date of the last modification.
+    :param ignore_if_range: If `False`, `If-Range` header will be taken into
+                            account.
     :return: `True` if the resource was modified, otherwise `False`.
     """
     if etag is None and data is not None:
@@ -792,18 +810,32 @@ def is_resource_modified(environ, etag=None, data=None, last_modified=None):
     if last_modified is not None:
         last_modified = last_modified.replace(microsecond=0)
 
-    modified_since = parse_date(environ.get('HTTP_IF_MODIFIED_SINCE'))
+    if_range = None
+    if not ignore_if_range and 'HTTP_RANGE' in environ:
+        # http://tools.ietf.org/html/rfc7233#section-3.2
+        # A server MUST ignore an If-Range header field received in a request
+        # that does not contain a Range header field.
+        if_range = parse_if_range_header(environ.get('HTTP_IF_RANGE'))
+
+    if if_range is not None and if_range.date is not None:
+        modified_since = if_range.date
+    else:
+        modified_since = parse_date(environ.get('HTTP_IF_MODIFIED_SINCE'))
 
     if modified_since and last_modified and last_modified <= modified_since:
         unmodified = True
+
     if etag:
-        if_none_match = parse_etags(environ.get('HTTP_IF_NONE_MATCH'))
-        if if_none_match:
-            # http://tools.ietf.org/html/rfc7232#section-3.2
-            # "A recipient MUST use the weak comparison function when comparing
-            # entity-tags for If-None-Match"
-            etag, _ = unquote_etag(etag)
-            unmodified = if_none_match.contains_weak(etag)
+        etag, _ = unquote_etag(etag)
+        if if_range is not None and if_range.etag is not None:
+            unmodified = parse_etags(if_range.etag).contains(etag)
+        else:
+            if_none_match = parse_etags(environ.get('HTTP_IF_NONE_MATCH'))
+            if if_none_match:
+                # http://tools.ietf.org/html/rfc7232#section-3.2
+                # "A recipient MUST use the weak comparison function when comparing
+                # entity-tags for If-None-Match"
+                unmodified = if_none_match.contains_weak(etag)
 
     return not unmodified
 
@@ -855,7 +887,7 @@ def is_hop_by_hop_header(header):
     .. versionadded:: 0.5
 
     :param header: the header to test.
-    :return: `True` if it's an entity header, `False` otherwise.
+    :return: `True` if it's an HTTP/1.1 "Hop-by-Hop" header, `False` otherwise.
     """
     return header.lower() in _hop_by_hop_headers
 
